"""
contains class for structured segmentgroups in mig tables. Builds table recursively.
"""

import json
from pathlib import Path
from types import NoneType
from typing import Any, Optional, Tuple

from loguru import logger
from maus.edifact import EdifactFormat
from pydantic import BaseModel, Field

from migmose.mig.nachrichtenstrukturtabelle import NachrichtenstrukturTabelle
from migmose.mig.nachrichtenstrukturzeile import NachrichtenstrukturZeile


class NestedNachrichtenstruktur(BaseModel):
    """
    class for structured segmentgroups in mig tables. Builds table recursively. Inherits from NachrichtenstrukturZeile
    e.g.(ORDCHG):
    {
    "segmente": [
        {
        "zaehler": "0160",
        "nr": "00007",
        "bezeichnung": "NAD",
        "standard_status": "M",
        "bdew_status": "M",
        "standard_maximale_wiederholungen": 1,
        "bdew_maximale_wiederholungen": 1,
        "ebene": 1,
        "inhalt": "MP-ID Absender"
        }
        ],
    "segmentgruppen": [
    {
        "segmente": [
        {
            "zaehler": "0260",
            "nr": "00008",
            "bezeichnung": "CTA",
            "standard_status": "M",
            "bdew_status": "M",
            "standard_maximale_wiederholungen": 1,
            "bdew_maximale_wiederholungen": 1,
            "ebene": 2,
            "inhalt": "Ansprechpartner"
            },
            {
            "zaehler": "0270",
            "nr": "00009",
            "bezeichnung": "COM",
            "standard_status": "C",
            "bdew_status": "R",
            "standard_maximale_wiederholungen": 5,
            "bdew_maximale_wiederholungen": 5,
            "ebene": 3,
            "inhalt": "Kommunikationsverbindung"
            }
            ],
        "segmentgruppen": []
        }
        ]
    }
    """

    header_linie: Optional[NachrichtenstrukturZeile] = None
    segmente: list[Optional[NachrichtenstrukturZeile]] = Field(default_factory=list)
    segmentgruppen: list[Optional["NestedNachrichtenstruktur"]] = Field(default_factory=list)

    @classmethod
    def create_nested_nachrichtenstruktur(
        cls, table: NachrichtenstrukturTabelle, header_line: Optional[NachrichtenstrukturZeile] = None, index: int = 0
    ) -> Tuple["NestedNachrichtenstruktur", int]:
        """init nested Nachrichtenstruktur"""
        collected_segments: list[Optional[NachrichtenstrukturZeile]] = []
        collected_segmentgroups: list[Optional["NestedNachrichtenstruktur"]] = []
        i = index
        while i < len(table.lines):
            line = table.lines[i]
            is_line_segmentgruppe = line.nr is None
            if is_line_segmentgruppe:
                added_segmentgroup, i = cls.create_nested_nachrichtenstruktur(table, line, i + 1)
                collected_segmentgroups.append(added_segmentgroup)
            else:
                collected_segments.append(line)
                i += 1
            if i < len(table.lines) and not isinstance(header_line, NoneType):
                is_next_line_segmentgruppe = table.lines[i].nr is None
                is_current_ebene_greater_than_next_ebene = line.ebene > table.lines[i].ebene
                is_current_header_ebene_greater_eq_than_next_ebene = header_line.ebene >= table.lines[i].ebene

                if (
                    is_next_line_segmentgruppe and is_current_header_ebene_greater_eq_than_next_ebene
                ) or is_current_ebene_greater_than_next_ebene:
                    return (
                        cls(
                            header_linie=header_line,
                            segmente=collected_segments,
                            segmentgruppen=collected_segmentgroups,
                        ),
                        i,
                    )
        return cls(header_linie=header_line, segmente=collected_segments, segmentgruppen=collected_segmentgroups), i

    def to_json(self, message_type: EdifactFormat, output_dir: Path) -> dict[str, Any]:
        """
        writes the NestedNachrichtenstruktur as json
        """
        output_dir.mkdir(parents=True, exist_ok=True)
        file_path = output_dir.joinpath("nested_nachrichtenstruktur.json")
        structured_json = self.model_dump()
        with open(file_path, "w", encoding="utf-8") as json_file:
            json.dump(structured_json, json_file, indent=4)
        logger.info(f"Wrote nested Nachrichtenstruktur for {message_type} to {file_path}")
        return structured_json



"""
contains class mig tables
"""

import csv
from pathlib import Path

from loguru import logger
from maus.edifact import EdifactFormat
from pydantic import BaseModel

from migmose.mig.nachrichtenstrukturzeile import NachrichtenstrukturZeile


class NachrichtenstrukturTabelle(BaseModel):
    """
    class for mig tables
    """

    lines: list[NachrichtenstrukturZeile]

    @classmethod
    def create_nachrichtenstruktur_tabelle(cls, raw_lines: list[str]) -> "NachrichtenstrukturTabelle":
        """
        reads table as list of raw lines and returns a NachrichtenstrukturTabelle
        consisting of NachrichtenstrukturZeilen
        """
        collected_lines: list[NachrichtenstrukturZeile] = []
        for raw_line in raw_lines:
            collected_lines.append(NachrichtenstrukturZeile.init_raw_lines(raw_line))
        return cls(lines=collected_lines)

    def to_csv(self, message_type: EdifactFormat, output_dir: Path) -> None:
        """
        writes the NestedNachrichtenstruktur as json
        """
        output_dir.mkdir(parents=True, exist_ok=True)
        file_path = output_dir.joinpath("nachrichtenstruktur.csv")
        fieldnames = list(NachrichtenstrukturZeile.model_fields.keys())
        with open(file_path, "w", encoding="utf-8", newline="") as csv_file:
            writer = csv.DictWriter(csv_file, fieldnames=fieldnames, dialect="excel", lineterminator="\n")
            writer.writeheader()
            for nachrichtenstruktur_zeile in self.lines:
                writer.writerow(nachrichtenstruktur_zeile.model_dump())
        logger.info(f"Wrote Nachrichtenstruktur for {message_type} to {file_path}")



"""
contains functions for file handling and parsing.
"""

import re
from collections import defaultdict
from datetime import datetime
from pathlib import Path
from typing import DefaultDict, Generator, Union

import click
import docx
from docx.document import Document
from docx.oxml import CT_Tbl  # type:ignore[attr-defined]
from docx.table import Table, _Cell
from loguru import logger
from maus.edifact import EdifactFormat, EdifactFormatVersion


def find_file_to_format(
    message_formats: list[EdifactFormat], edi_energy_repo: Path, format_version: EdifactFormatVersion
) -> dict[EdifactFormat, Path]:
    """
    finds the file with the message type in the input directory
    """
    input_dir = edi_energy_repo / Path("edi_energy_de") / Path(format_version)
    all_file_dict: DefaultDict[EdifactFormat, list[Path]] = defaultdict(list)
    file_dict: dict[EdifactFormat, Path] = {}
    for message_format in message_formats:
        for file in input_dir.iterdir():
            if "MIG" not in file.name or file.suffix != ".docx":
                continue
            if message_format is EdifactFormat.UTILMDG and "Gas" in file.name:
                all_file_dict[EdifactFormat.UTILMDG].append(file)
            elif message_format is EdifactFormat.UTILMDS and "Strom" in file.name:
                all_file_dict[EdifactFormat.UTILMDS].append(file)
            elif message_format in file.name:
                all_file_dict[message_format].append(file)
        if len(all_file_dict[message_format]) == 0:
            logger.warning(f"⚠️ No file found for {message_format}", fg="red")
            continue
        file_dict[message_format] = get_latest_file(all_file_dict[message_format])
    if file_dict:
        return file_dict
    logger.error("❌ No files found in the input directory.", fg="red")
    raise click.Abort()


def _extract_date(file_path: Path) -> tuple[datetime, Path]:
    # Regex to extract the date format YYYYMMDD from the filename as a string
    match = re.search(r"(\d{8})\.docx$", file_path.name)
    if match:
        # Return the date as a datetime object for comparison and the path for use
        return datetime.strptime(match.group(1), "%Y%m%d"), file_path
    logger.warning(
        f"⚠️ No timestamp in filename found in {file_path}."
        + "in case of multiple docx files in this path, it must be a "
        + "timestamp with format 'yyyyMMdd.docx' in filename.",
        fg="red",
    )
    raise click.Abort()


def get_latest_file(file_list: list[Path]) -> Path:
    """
    This function takes a list of docx files Path
    and returns the Path of the latest MIG docx file based on the timestamp in its name.
    The timestamp is assumed to be formatted as YYYYMMDD and located just before the ".docx" extension.

    Parameters:
        file_list (list of Path): A list containing file paths with timestamps.

    Returns:
        Path: The path of the latest file. Returns None if no valid date is found.
    """
    # Initialize variables to keep track of the latest file and date
    latest_file: Path
    latest_date: datetime | None = None

    for file_path in file_list:
        date, path = _extract_date(file_path)
        if latest_date is None or date > latest_date:
            latest_file = path
            latest_date = date

    # Return the path of the file with the latest date
    return latest_file


def get_paragraphs_up_to_diagram(parent: Union[Document, _Cell]) -> Generator[Table, None, None]:
    """Goes through paragraphs and tables"""
    # pylint: disable=protected-access
    if isinstance(parent, Document):
        parent_elm = parent.element.body
    elif isinstance(parent, _Cell):
        parent_elm = parent._tc
    else:
        raise ValueError("Passed parent argument must be of type Document or _Cell")

    for child in parent_elm.iterchildren():
        if isinstance(child, CT_Tbl):
            yield Table(child, parent)


_row_regex = re.compile(r"^(?P<left>\t\d+\t)(?P<nr>\d{0,5})(?P<right>\t.*)$")
"""
https://regex101.com/r/vtF07B/2
"""


def _zfill_nr(row_str: str) -> str:
    match = _row_regex.match(row_str)
    if not match:
        return row_str
    left = match.group("left")
    nr = match.group("nr")
    right = match.group("right")
    return f"{left}{nr.zfill(5)}{right}"


def parse_raw_nachrichtenstrukturzeile(input_path: Path) -> list[str]:
    """
    parses raw nachrichtenstrukturzeile from a table. returns list of raw lines
    """
    # pylint: disable=protected-access
    doc = docx.Document(str(input_path.absolute()))
    docx_objects = get_paragraphs_up_to_diagram(doc)
    mig_tables = []
    nachrichtenstruktur_header = "Status\tMaxWdh\n\tZähler\tNr\tBez\tSta\tBDEW\tSta\tBDEW\tEbene\tInhalt"
    for docx_object in docx_objects:
        for ind, line in enumerate(docx_object._cells):
            # marks the beginning of the complete nachrichtentruktur table
            if line.text == nachrichtenstruktur_header:
                mig_tables.extend([row.text for row in docx_object._cells[ind + 1 :]])
                break

    # filter empty rows and headers
    mig_tables = [_zfill_nr(row) for row in mig_tables if row not in ("", "\n", nachrichtenstruktur_header)]
    return mig_tables


def _extract_document_version(path: Path) -> str:
    document_str = str(path)
    pattern = (
        r"MIG(?:Strom|Gas)?-?informatorischeLesefassung?(.*?)"
        r"(?:_|KonsolidierteLesefassung|-AußerordentlicheVeröffentlichung)"
    )
    matches = re.search(pattern, document_str, re.IGNORECASE)
    if matches:
        document_version = matches.group(1)
        if document_version == "":
            logger.warning(f"❌ No document version found in {path}.", fg="red")
        return document_version
    logger.error(f"❌ Unexpected document name in {path}.", fg="red")
    return ""



"""
contains class for trees consisting of segments of mig tables
"""

import json
from pathlib import Path
from typing import Optional

from loguru import logger
from maus.edifact import EdifactFormat
from pydantic import BaseModel, Field

from migmose.mig.reducednestednachrichtenstruktur import ReducedNestedNachrichtenstruktur


class SegmentGroupHierarchy(BaseModel):
    """
    This module contains the same model for a segment group hierarchy as used by the MAUS library,
    which can be found here:
    (https://github.com/Hochfrequenz/mig_ahb_utility_stack/blob/eae2d84d600f42257cf364190da486aa011a5882/src/maus/models/message_implementation_guide.py#L17). # pylint: disable=line-too-long
    However, instead of using the attrs class, we are creating a Pydantic class.
    """

    opening_segment: Optional[str] = None
    segment_group: Optional[str] = None
    sub_hierarchy: list[Optional["SegmentGroupHierarchy"]] = Field(default_factory=list)

    @classmethod
    def create_segmentgroup_hierarchy(
        cls, reduced_nested_nachrichtenstruktur: ReducedNestedNachrichtenstruktur
    ) -> "SegmentGroupHierarchy":
        """init Segmentrgroup Hierarchy"""
        segment_group: Optional[str] = None
        opening_segment: Optional[str] = None
        if reduced_nested_nachrichtenstruktur.header_linie is not None:
            segment_group = reduced_nested_nachrichtenstruktur.header_linie.bezeichnung
        if (
            any(reduced_nested_nachrichtenstruktur.segmente)
            and reduced_nested_nachrichtenstruktur.segmente[0] is not None
        ):
            opening_segment = reduced_nested_nachrichtenstruktur.segmente[0].bezeichnung
        sub_hierarchy: list[Optional["SegmentGroupHierarchy"]] = []
        for sub_segmentgroup in reduced_nested_nachrichtenstruktur.segmentgruppen:
            if sub_segmentgroup is not None:
                sub_hierarchy.append(cls.create_segmentgroup_hierarchy(sub_segmentgroup))

        return SegmentGroupHierarchy(
            segment_group=segment_group, opening_segment=opening_segment, sub_hierarchy=sub_hierarchy
        )

    def to_json(self, message_type: EdifactFormat, output_dir: Path) -> None:
        """
        writes the reduced NestedNachrichtenstruktur as json
        """
        output_dir.mkdir(parents=True, exist_ok=True)
        file_path = output_dir.joinpath(f"{message_type}.sgh.json")
        structured_json = self.model_dump(mode="json")
        with open(file_path, "w", encoding="utf-8") as json_file:
            json.dump(structured_json, json_file, indent=4)
        logger.info("Wrote segmentgroup hierarchy {}.(sgh.json) for {} to {}", message_type, message_type, file_path)



"""
contains CLI logic for migmose.
"""

from pathlib import Path

import click
from loguru import logger
from maus.edifact import EdifactFormat, EdifactFormatVersion

from migmose.mig.nachrichtenstrukturtabelle import NachrichtenstrukturTabelle
from migmose.mig.nestednachrichtenstruktur import NestedNachrichtenstruktur
from migmose.mig.reducednestednachrichtenstruktur import ReducedNestedNachrichtenstruktur
from migmose.mig.segmentgrouphierarchies import SegmentGroupHierarchy
from migmose.parsing import _extract_document_version, find_file_to_format, parse_raw_nachrichtenstrukturzeile


def check_message_format(ctx, param, value) -> list[EdifactFormat]:  # type: ignore[no-untyped-def] # pylint: disable=unused-argument
    """
    Check if the message format is valid.
    """
    if value is None:
        value = map(lambda x: x.name, EdifactFormat)
    return list(value)


# add CLI logic
@click.command()
@click.option(
    "-eemp",
    "--edi-energy-mirror-path",
    type=click.Path(exists=True, file_okay=False, dir_okay=True, resolve_path=True, path_type=Path),
    prompt="Please enter the path to your local edi energy mirror repository.",
    help="The root path to the edi_energy_mirror repository.",
    required=True,
)
@click.option(
    "-mf",
    "--message-format",
    type=click.Choice(list(map(lambda x: x.name, EdifactFormat)), case_sensitive=False),
    # Taken from https://github.com/pallets/click/issues/605#issuecomment-889462570
    help="Defines the set of message formats to be parsed. If no format is specified, all formats are parsed.",
    multiple=True,
    callback=check_message_format,
)
@click.option(
    "-fv",
    "--format-version",
    multiple=False,
    type=click.Choice([e.value for e in EdifactFormatVersion], case_sensitive=False),
    default=EdifactFormatVersion.FV2404,
    required=True,
    help="Format version of the MIG documents, e.g. FV2310",
)
@click.option(
    "-o",
    "--output-dir",
    type=click.Path(exists=False, dir_okay=True, file_okay=False, path_type=Path),
    prompt="Please enter the path to the directory which should contain the output files.",
    help="Set path to directory which contains the output files. If the directory does not exist, it will be created.",
)
@click.option(
    "-ft",
    "--file-type",
    type=click.Choice(["csv", "nested_json", "reduced_nested_json", "tree"], case_sensitive=False),
    default=["csv", "nested_json", "reduced_nested_json", "tree"],
    help="Defines the output format. Choose between csv and nested_json and reduced_nested_json. Default is csv.",
    multiple=True,
)
def main(
    edi_energy_mirror_path: Path,
    output_dir: Path,
    format_version: EdifactFormatVersion | str,
    message_format: list[EdifactFormat],
    file_type: tuple[str],
) -> None:
    """
    Main function. Uses CLI input.
    """
    if isinstance(format_version, str):
        format_version = EdifactFormatVersion(format_version)
    if int(str(format_version)[-4:]) >= 2310 and EdifactFormat.UTILMD in message_format:
        logger.warning(
            "💡 UTILMD is not available for format versions 2310 and above. Parse UTILMDS and UTILMDG instead."
        )
        message_format.remove(EdifactFormat.UTILMD)
        message_format.extend([EdifactFormat.UTILMDG, EdifactFormat.UTILMDS])
    dict_files = find_file_to_format(message_format, edi_energy_mirror_path, format_version)
    for m_format, file in dict_files.items():
        output_dir_for_format = output_dir / format_version / m_format
        raw_lines = parse_raw_nachrichtenstrukturzeile(file)
        nachrichtenstrukturtabelle = NachrichtenstrukturTabelle.create_nachrichtenstruktur_tabelle(raw_lines)
        nested_nachrichtenstruktur, _ = NestedNachrichtenstruktur.create_nested_nachrichtenstruktur(
            nachrichtenstrukturtabelle
        )
        reduced_nested_nachrichtenstruktur = ReducedNestedNachrichtenstruktur.create_reduced_nested_nachrichtenstruktur(
            nested_nachrichtenstruktur
        )
        if "csv" in file_type:
            logger.info(
                "💾 Saving flat Nachrichtenstruktur table for {} and {} as csv to {}.",
                m_format,
                format_version,
                output_dir_for_format,
            )
            nachrichtenstrukturtabelle.to_csv(m_format, output_dir_for_format)
        if "nested_json" in file_type:
            # Save the nested Nachrichtenstruktur as json
            logger.info(
                "💾 Saving nested Nachrichtenstruktur for {} and {} as json to {}.",
                m_format,
                format_version,
                output_dir_for_format,
            )
            nested_nachrichtenstruktur.to_json(m_format, output_dir_for_format)

        if "reduced_nested_json" in file_type:
            # Save the reduced nested Nachrichtenstruktur as json
            logger.info(
                "💾 Saving reduced nested Nachrichtenstruktur for {} and {} as json to {}.",
                m_format,
                format_version,
                output_dir_for_format,
            )
            reduced_nested_nachrichtenstruktur.to_json(m_format, output_dir_for_format)
        if "sgh_json" in file_type:
            sgh = SegmentGroupHierarchy.create_segmentgroup_hierarchy(reduced_nested_nachrichtenstruktur)
            # Save the reduced nested Nachrichtenstruktur as json
            logger.info(
                "💾 Saving reduced nested Nachrichtenstruktur for {} and {} as json to {}.",
                m_format,
                format_version,
                output_dir_for_format,
            )
            sgh.to_json(m_format, output_dir_for_format)
        if "tree" in file_type:
            # Save the reduced nested Nachrichtenstruktur as json
            logger.info(
                "💾 Saving tree for {} and {} as json to {}.",
                m_format,
                format_version,
                output_dir_for_format,
            )
            document_version = _extract_document_version(file)
            reduced_nested_nachrichtenstruktur.output_tree(m_format, output_dir_for_format, document_version)


if __name__ == "__main__":
    main()  # pylint:disable=no-value-for-parameter



"""
contains class for trees consisting of segments of mig tables
"""

import json
from collections import defaultdict
from pathlib import Path
from typing import Any, DefaultDict, Optional, TypeAlias

from jinja2 import Template
from loguru import logger
from maus.edifact import EdifactFormat
from pydantic import BaseModel, Field

from migmose.mig.nachrichtenstrukturzeile import NachrichtenstrukturZeile
from migmose.mig.nestednachrichtenstruktur import NestedNachrichtenstruktur

_SegmentDict: TypeAlias = (
    dict[
        tuple[str, str],
        tuple[
            list[Optional[NachrichtenstrukturZeile]],
            Optional[NachrichtenstrukturZeile],
            set[tuple[str, str]],
        ],
    ]
    | None
)


# Helper function to create a unique identifier for each segment
def _get_identifier(segment: Optional[NachrichtenstrukturZeile]) -> tuple[str, str]:
    if segment is None:
        return "0", "root"
    return segment.zaehler, segment.bezeichnung


# Function to process segments and remove duplicates within the same list.
def _process_segments(
    segments: list[Optional[NachrichtenstrukturZeile]],
) -> list[Optional[NachrichtenstrukturZeile]]:
    seen = set()
    unique_segments: list[Optional[NachrichtenstrukturZeile]] = []
    for segment in segments:
        if segment is not None:
            identifier = _get_identifier(segment)
            if identifier not in seen:
                seen.add(identifier)
                unique_segments.append(segment)
    return unique_segments


# Recursive function to traverse and clean segment groups
def _process_segmentgruppen(
    segmentgruppen_identifiers: set[tuple[str, str]],
    segment_dict: _SegmentDict,
    depth: int = 0,
) -> list[Optional["ReducedNestedNachrichtenstruktur"]]:
    """
    Recursively clean segment groups to ensure nested nachrichtenstruktur consisting only of a unique subset.
    """
    result: list[Optional[ReducedNestedNachrichtenstruktur]] = []

    for sg in sorted(segmentgruppen_identifiers):
        if sg is not None:
            # not sure about those type hints... they please mypy but I'm not the dev of this code
            segmente: list[Optional[NachrichtenstrukturZeile]]
            header_line: Optional[NachrichtenstrukturZeile]
            segmentgroups: set[tuple[str, str]]
            segmente, header_line, segmentgroups = segment_dict[sg]  # type:ignore[index]
            if segmente is not None:
                segmente = sorted(segmente, key=lambda x: x.zaehler)
            _new_sg = ReducedNestedNachrichtenstruktur(header_linie=header_line, segmente=segmente)
            _new_sg.segmentgruppen = _process_segmentgruppen(segmentgroups, segment_dict, depth + 1)
            result.append(_new_sg)
            logger.info("Added {} with {} segments at depth {}.", sg, len(segmente), depth)
    return result


def _build_segment_dict(
    segment_groups: list[Optional[NestedNachrichtenstruktur]],
    segment_dict: _SegmentDict = None,
) -> dict[
    tuple[str, str],
    tuple[list[Optional[NachrichtenstrukturZeile]], Optional[NachrichtenstrukturZeile], set[tuple[str, str]]],
]:
    """Build a dictionary of segments and segment groups to find unique set."""
    if segment_dict is None:
        segment_dict = {}
    for _sg in segment_groups:
        if _sg is not None:
            name = _get_identifier(_sg.header_linie)

            # Check if the current segments are already known and complete by unknown segments
            if name in segment_dict:
                # make sure every possible segment is included
                segment_dict[name] = (
                    _process_segments(_sg.segmente + segment_dict[name][0]),
                    segment_dict[name][1],
                    segment_dict[name][2],
                )
            else:
                segment_dict[name] = (_process_segments(_sg.segmente), _sg.header_linie, set())

            # Iterate recursively through nested segmentgroups
            for segmentgruppe in _sg.segmentgruppen:
                if segmentgruppe is not None:
                    sg_name = _get_identifier(segmentgruppe.header_linie)
                    segment_dict[name][2].add(sg_name)
                    segment_dict = _build_segment_dict([segmentgruppe], segment_dict)

    return segment_dict


def _dict_to_tree_str(tree: DefaultDict[str, list[NachrichtenstrukturZeile]]) -> str:
    template_str = """{%- for key, segment_list in tree.items() -%}
    {{-key-}}:{%- for segment in segment_list -%}
    {{segment.bezeichnung}}[{{segment.standard_status}};{{segment.bdew_status}}]
    {%- if not loop.last -%},{%- endif -%}
    {%- endfor -%}{{"\n"}}
    {%- endfor -%}
    """
    template = Template(template_str)
    return template.render(tree=tree)


def _build_tree_dict(
    reduced_nestednachrichtenstruktur: "ReducedNestedNachrichtenstruktur",
    tree_dict: Optional[DefaultDict[str, list[NachrichtenstrukturZeile]]] = None,
) -> DefaultDict[str, list[NachrichtenstrukturZeile]]:
    """
    Build a dictionary to compose the .tree files in the MAUS library.
    """
    if tree_dict is None:
        tree_dict = defaultdict(list)
    if (
        reduced_nestednachrichtenstruktur.header_linie is None
        and reduced_nestednachrichtenstruktur.segmente[0] is not None
    ):
        tree_dict["/"] = [
            NachrichtenstrukturZeile(
                zaehler="0",
                nr="00000",
                bezeichnung="UNB",
                standard_status="M",
                bdew_status="M",
                standard_maximale_wiederholungen=0,
                bdew_maximale_wiederholungen=0,
                ebene=0,
                inhalt="0",
            ),
            reduced_nestednachrichtenstruktur.segmente[0],
            NachrichtenstrukturZeile(
                zaehler="0",
                nr="00000",
                bezeichnung="UNZ",
                standard_status="M",
                bdew_status="M",
                standard_maximale_wiederholungen=0,
                bdew_maximale_wiederholungen=0,
                ebene=0,
                inhalt="0",
            ),
        ]
        tree_dict["UNH"].extend(
            [
                segment
                for segment in reduced_nestednachrichtenstruktur.segmente
                if segment and segment.bezeichnung not in ["UNH", "UNT"]
            ]
        )
        tree_dict["UNH"].extend(
            [sg.header_linie for sg in reduced_nestednachrichtenstruktur.segmentgruppen if sg and sg.header_linie]
        )
        tree_dict["UNH"].extend(
            [
                segment
                for segment in reduced_nestednachrichtenstruktur.segmente
                if segment and segment.bezeichnung in ["UNT"]
            ]
        )
    elif reduced_nestednachrichtenstruktur.header_linie is not None:
        if reduced_nestednachrichtenstruktur.segmente not in [[], [None]]:
            tree_dict[reduced_nestednachrichtenstruktur.header_linie.bezeichnung].extend(
                [segment for segment in reduced_nestednachrichtenstruktur.segmente if segment]
            )
        if reduced_nestednachrichtenstruktur.segmentgruppen:
            tree_dict[reduced_nestednachrichtenstruktur.header_linie.bezeichnung].extend(
                [sg.header_linie for sg in reduced_nestednachrichtenstruktur.segmentgruppen if sg and sg.header_linie]
            )
    else:
        raise ValueError("No header line or segment found.")
    for segmentgruppe in reduced_nestednachrichtenstruktur.segmentgruppen:
        if segmentgruppe is not None:
            tree_dict = _build_tree_dict(segmentgruppe, tree_dict)
    return tree_dict


class ReducedNestedNachrichtenstruktur(BaseModel):
    """will contain the tree structure of nachrichtenstruktur tables"""

    header_linie: Optional[NachrichtenstrukturZeile] = None
    segmente: list[Optional[NachrichtenstrukturZeile]] = Field(default_factory=list)
    segmentgruppen: list[Optional["ReducedNestedNachrichtenstruktur"]] = Field(default_factory=list)

    @classmethod
    def create_reduced_nested_nachrichtenstruktur(
        cls, nachrichten_struktur: NestedNachrichtenstruktur
    ) -> "ReducedNestedNachrichtenstruktur":
        """init nested Nachrichtenstruktur"""

        data = ReducedNestedNachrichtenstruktur()
        # Start processing the top-level segments
        if nachrichten_struktur.segmente is not None:
            data.segmente = _process_segments(nachrichten_struktur.segmente)

        # Process segment groups recursively
        if nachrichten_struktur.segmentgruppen is not None:
            segment_dict = _build_segment_dict(nachrichten_struktur.segmentgruppen)
            segmentgruppen_identifiers = set(
                _get_identifier(sg.header_linie) for sg in nachrichten_struktur.segmentgruppen if sg is not None
            )
            data.segmentgruppen = _process_segmentgruppen(segmentgruppen_identifiers, segment_dict)

        return data

    def to_json(self, message_type: EdifactFormat, output_dir: Path) -> dict[str, Any]:
        """
        writes the reduced NestedNachrichtenstruktur as json
        """
        output_dir.mkdir(parents=True, exist_ok=True)
        file_path = output_dir.joinpath("reduced_nested_nachrichtenstruktur.json")
        structured_json = self.model_dump()
        with open(file_path, "w", encoding="utf-8") as json_file:
            json.dump(structured_json, json_file, indent=4)
        logger.info("Wrote reduced nested Nachrichtenstruktur for {} to {}", message_type, file_path)
        return structured_json

    def output_tree(self, message_type: EdifactFormat, output_dir: Path, document_version: str) -> None:
        """Writes reduced NestedNachrichtenstruktur in the .tree grammar of MAUS."""
        # generate tree dict
        tree_dict = _build_tree_dict(self)
        # convert tree dict to string
        tree_str = _dict_to_tree_str(tree_dict)
        # write tree file
        output_dir.mkdir(parents=True, exist_ok=True)
        file_path = output_dir / f"{message_type}{document_version}.tree"
        with open(file_path, "w", encoding="utf-8") as tree_file:
            tree_file.write(tree_str)
        logger.info("Wrote reduced .tree file for {} to {}", message_type, file_path)



"""
contains class for lines in mig tables
"""

from typing import Any

from pydantic import BaseModel, Field


class NachrichtenstrukturZeile(BaseModel):
    """
    class for lines in mig tables, e.g. (ORDCHG):
    {
        "zaehler": "0010",
        "nr": "00001",
        "bezeichnung": "UNH",
        "standard_status": "M",
        "bdew_status": "M",
        "standard_maximale_wiederholungen": 1,
        "bdew_maximale_wiederholungen": 1,
        "ebene": 0,
        "inhalt": "Nachrichten-Kopfsegment"
        }
    """

    zaehler: str
    nr: str | None = Field(default=None, pattern=r"^\d{5}$")  #: the segment ID (can be used to match with AHBs)
    bezeichnung: str
    standard_status: str
    bdew_status: str
    standard_maximale_wiederholungen: int
    bdew_maximale_wiederholungen: int
    ebene: int
    inhalt: str

    @classmethod
    def init_raw_lines(cls, raw_line: str) -> "NachrichtenstrukturZeile":
        """
        reads one raw line and returns a NachrichtenstrukturZeile object
        """
        fields = raw_line.split("\t")[1:]
        field_names = [
            "zaehler",
            "nr",
            "bezeichnung",
            "standard_status",
            "bdew_status",
            "standard_maximale_wiederholungen",
            "bdew_maximale_wiederholungen",
            "ebene",
            "inhalt",
        ]
        is_line_segmentgroup = len(fields) == len(field_names) - 1
        is_line_incomplete = len(fields) != len(field_names) and not is_line_segmentgroup

        if is_line_segmentgroup:
            field_names = field_names[:1] + field_names[2:]
        if is_line_incomplete:
            raise ValueError(f"Expected 8 or 9 fields, got {len(fields)}, line: {raw_line}")
        field_dict: dict[str, Any] = dict(zip(field_names, fields))
        if "nr" in field_dict and len(field_dict["nr"]) < 5:
            field_dict["nr"] = field_dict["nr"].zfill(5)
        return cls(**field_dict)
